#ifndef DUNE_ASH_GLUE_L2PROJECTION_HH
#define DUNE_ASH_GLUE_L2PROJECTION_HH

#include <dune/geometry/quadraturerules.hh>

#include <dune/fem/common/bindguard.hh>
#include <dune/fem/function/localfunction/const.hh>
#include <dune/fem/function/localfunction/temporary.hh>
#include <dune/fem/operator/1order/localmassmatrix.hh>
#include <dune/fem/quadrature/elementquadrature.hh>

namespace Dune
{

  namespace Ash
  {

    template< class GridPartGlue >
    class GluedL2Projection
    {
      typedef GluedL2Projection< GridPartGlue > ThisType;

    public:
      typedef GridPartGlue GridPartGlueType;

    private:
      typedef typename GridPartGlueType::ctype ctype;

      static const int dimension = GridPartGlueType::dimension;

      typedef typename GridPartGlueType::IntersectionIterator IntersectionIterator;
      typedef typename GridPartGlueType::Intersection Intersection;

      typedef typename GridPartGlueType::InsideGridPartType DomainGridPart;
      typedef typename GridPartGlueType::OutsideGridPartType RangeGridPart;

      typedef typename Intersection::InsideEntity DomainEntity;
      typedef typename Intersection::OutsideEntity RangeEntity;

      typedef Dune::QuadratureRules< ctype, dimension > QuadratureRules;

    public:
      explicit GluedL2Projection ( const GridPartGlueType &glue )
        : glue_( glue )
      {}

      template< class DomainFunction, class RangeFunction >
      void operator() ( const DomainFunction &u, RangeFunction &w ) const
      {
        typedef typename RangeFunction::DiscreteFunctionSpaceType RangeSpace;

        typedef Fem::LocalMassMatrix< RangeSpace, Fem::ElementQuadrature< RangeGridPart, 0 > > LocalMassMatrix;

        // assemle right hand side into w

        w.clear();

        Fem::ConstLocalFunction< DomainFunction > uLocal( u );
        Fem::TemporaryLocalFunction< RangeSpace > wLocal( w.space() );

        for( const Intersection &intersection : glue_ )
        {
          const typename Intersection::Geometry geometry = intersection.geometry();

          const DomainEntity inside = intersection.inside();
          auto uGuard = Fem::bindGuard( uLocal, inside );

          const RangeEntity outside = intersection.outside();
          wLocal.init( outside );
          wLocal.clear();

          const int order = uLocal.order() + wLocal.order();
          for( const auto qp : QuadratureRules::rule( intersection.type(), order ) )
          {
            auto value = uLocal.evaluate( intersection.geometryInInside().global( qp.position() ) );
            value *= qp.weight() * geometry.integrationElement( qp.position() );
            wLocal.axpy( intersection.geometryInOutside().global( qp.position() ), value );
          }

          w.addLocalDofs( outside, wLocal.localDofVector() );
        }

        // apply inverse mass matrix to w (in place)

        LocalMassMatrix localMassMatrix( w.space(), 2*w.space().order() );

        for( const RangeEntity &entity : w.space() )
        {
          wLocal.init( entity );
          w.getLocalDofs( entity, wLocal.localDofVector() );
          localMassMatrix.applyInverse( entity, wLocal );
          w.setLocalDofs( entity, wLocal.localDofVector() );
        }
      }

    private:
      const GridPartGlueType &glue_;
    };

  } // namespace Ash

} // namespace Dune

#endif // #ifndef DUNE_ASH_GLUE_L2PROJECTION_HH
