// vim: set expandtab ts=2 sw=2 sts=2: 
#ifndef VECTORFIELDVISUALIZER_HH
#define VECTORFIELDVISUALIZER_HH

#include <cmath>
#include <iostream>
#include <string>
#include <vector>

#include <dune/common/fvector.hh>

#include <dune/grid/common/rangegenerators.hh>

#include <dune/fem/io/parameter.hh>

#include <dune/ash/opengl/displaylist.hh>
#include <dune/ash/opengl/opengl.hh>

namespace Dune
{

  namespace Ash
  {

    // VectorFieldVisualizer
    // ---------------------

    class VectorFieldVisualizerBase
    {
      typedef VectorFieldVisualizerBase ThisType;

    public:
      typedef FieldVector< GLfloat, 2 > VectorType;

      explicit VectorFieldVisualizerBase ( const Dune::Ash::DisplayList &arrowList )
      : arrowList_( &arrowList )
      {}

      // 'v' is the value at the center of the cell
      void drawArrow ( const VectorType &centerCoordinate,
                       const VectorType &direction,
                       GLfloat scalingFactor,
                       GLfloat colorScaling ) const
      {
        const GLfloat abs = direction.two_norm();
        GLfloat v[ 4 ] = { (direction[ 0 ] / abs), (direction[ 1 ] / abs) };

        GLfloat rotationMatrix[ 16 ] = { v[ 0 ], -v[ 1 ], 0.0f, 0.0f,
                                         v[ 1 ],  v[ 0 ], 0.0f, 0.0f,
                                         0.0f,    0.0f,   1.0f, 0.0f,
                                         0.0f,    0.0f,   0.0f, 1.0f };

        const GLfloat redOffset = std::min( std::max( abs / colorScaling, 0.0f ), 1.0f );

        glColor( 0.0f, 0.0f, 1.0f, redOffset );

        glMatrixMode( GL_MODELVIEW );
        glPushMatrix();
        glTranslate( centerCoordinate[ 0 ], 1.0f - centerCoordinate[ 1 ], 0.0f );
        glMultMatrixf( rotationMatrix );
        glScalef( scalingFactor, scalingFactor, 0. );
      
        arrowList()();
        glPopMatrix();
      }

    private:
      const Dune::Ash::DisplayList &arrowList () const { return *arrowList_; }

      const Dune::Ash::DisplayList *arrowList_;
    };



    // VectorFieldVisualizer
    // ---------------------

    template< typename DiscreteFunction >
    class VectorFieldVisualizer
      : public VectorFieldVisualizerBase
    {
      typedef VectorFieldVisualizer< DiscreteFunction > ThisType;
      typedef VectorFieldVisualizerBase BaseType;

    public:
      typedef DiscreteFunction DiscreteFunctionType;
      typedef typename DiscreteFunctionType::DiscreteFunctionSpaceType DiscreteFunctionSpaceType;

    protected:
      using BaseType::drawArrow;

    public:
      VectorFieldVisualizer ( const Dune::Ash::DisplayList &arrowList, const DiscreteFunctionType &discreteFunction )
        : BaseType( arrowList ), discreteFunction_( discreteFunction )
      {}

      VectorFieldVisualizer ( const ThisType & ) = delete;
      ThisType &operator= ( const ThisType & ) = delete;

      void operator() ( double colorScaling ) const
      {
        Fem::ConstLocalFunction< DiscreteFunctionType > local( discreteFunction_ );
        for( const auto &entity : elements( discreteFunction_.gridPart(), Partitions::interiorBorder ) )
        {
          local.init( entity );
          const auto geometry = entity.geometry();

          decltype( geometry.local( geometry.center() ) ) x( 0.5 );
          drawArrow( geometry.center(), local.evaluate( x ),
                     static_cast< GLfloat >( geometry.corner( 1 )[ 0 ] - geometry.corner( 0 )[ 0 ] ),
                     static_cast< GLfloat >( colorScaling ) );
        }
      }

    private:
      const DiscreteFunctionType &discreteFunction_;
    };

  } // namespace Ash

} // namespace Dune

#endif // #ifndef VECTORFIELDVISUALIZER_HH
